const LATEST_VIEW = 0;
const CHANGES_VIEW = 1;
const HISTORY_VIEW = 2;

function ModelTreeData(rdfservice, rdfcu, builder, view)
{
  this.rdfservice = rdfservice;
  this.rdfcu = rdfcu;
  this.builder = builder;
  this.view = view;
  // Set up some RDF resources...
  this.rr =
  {
    root: this.rdfservice.GetResource(
      "http://www.cellml.org/tools/pcenv/all-models"),
    name: this.rdfservice.GetResource(
      "http://www.cellml.org/tools/pcenv/models#Name"),
    open: this.rdfservice.GetResource(
      "http://home.netscape.com/NC-rdf#open"),
    true: this.rdfservice.GetLiteral("true")
  }

  this.persist = this.rdfservice.GetDataSource("rdf:local-store");

  var c = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-" +
                             "datasource"];
  var ifce = Components.interfaces.nsIRDFDataSource;


  // Create three new datasources, one for each view...
  this.sources = [];
  this.rdfroot = [];

  this.nodeByResource = {};
  this.resourceByNode = {};
  this.latestResourceByTopNode = {};

  var i;
  for (i = 0; i < 3; i++)
  {
    this.sources[i] = c.createInstance(ifce);
    // Create a top-level container...
    this.rdfroot[i] = this.rdfcu.MakeSeq(this.sources[i], this.rr.root);
  }

  // It is now time to recursively set up the RDF sources from the context...
  var lmRoot = window.context.loadedModels;
  var sourceStack = [lmRoot];
  // We do a depth-first search, building the CHANGES_VIEW RDF, and
  // making lists in preparation for HISTORY_VIEW and LATEST_VIEW.
  var propStack = [[this.rr.root, this.rdfroot[CHANGES_VIEW], null]];
  this.rdfChangeSequenceByList = {};
  this.rdfChangeSequenceByList[
    lmRoot.QueryInterface(Components.interfaces.IWrappedPCM).objid] =
    this.rdfroot[CHANGES_VIEW];
  var latestlist = [];
  var historylist = [];

  while (sourceStack.length)
  {
    var ssi = sourceStack.pop();
    var props = propStack.pop();
    var it = ssi.iterateModelNodes();

    while (true)
    {
      var n = it.nextModelNode();
      if (n == null)
        break;
      var l = n.derivedModels;
      sourceStack.push(l);
      var rsrc = this.rdfservice.GetAnonymousResource();
      this.persist.Assert(rsrc, this.rr.open, this.rr.true, true);
      this.nodeByResource[rsrc.ValueUTF8] = n;
      this.resourceByNode[n.QueryInterface(Components.interfaces.
                                           IWrappedPCM).objid] = rsrc;
      var seq = this.rdfcu.MakeSeq(this.sources[CHANGES_VIEW], rsrc);
      this.rdfChangeSequenceByList[
        l.QueryInterface(Components.interfaces.IWrappedPCM).objid] = seq;
      var latestdata = props[2];
      var mt = n.modificationTimestamp;
      if (latestdata == null)
      {
        // Note: Don't update latestdata[3], it is supposed to be stuck as the
        // toplevel node corresponding to the latest, not the latest.
        // This differs from latestdata[0], which gets updated.
        latestdata = [n, mt, rsrc, n];
        latestlist.push(latestdata);
      }
      else if (latestdata[1] < mt)
      {
        latestdata[0] = n;
        latestdata[1] = mt;
        latestdata[2] = rsrc;
      }
      // Add a record in the history list.
      historylist.push([n, mt, rsrc]);

      // Add the element...
      props[1].AppendElement(rsrc);
      // Set the name...
      var i;
      var l = this.rdfservice.GetLiteral(n.name);
      for (i = 0; i < 3; i++)
        this.sources[i].Assert(rsrc, this.rr.name, l, true);

      // Set up propstack for the next trip...
      propStack.push([rsrc, seq, latestdata]);
    }
  }
  // Sort the history list...
  historylist.sort(function(a,b) {return (a[1] - b[1]); });

  this.historySeq = this.rdfcu.MakeSeq(this.sources[HISTORY_VIEW],
                                       this.rr.root);
  for (record in historylist)
  {
    var record = historylist[record];
    var rsrc = record[2];
    this.historySeq.AppendElement(rsrc);
  }

  this.latestSeq = this.rdfcu.MakeSeq(this.sources[LATEST_VIEW],
                                      this.rr.root);
  for (record in latestlist)
  {
    var record = latestlist[record];
    var rsrc = record[2];
    this.latestSeq.AppendElement(rsrc);
    var objid = record[3].QueryInterface(Components.interfaces.IWrappedPCM).
      objid;
    this.latestResourceByTopNode[objid] = rsrc;
  }
}

ModelTreeData.prototype =
{
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Components.interfaces.cellml_contextIModelNodeMonitor) ||
        aIID.equals(Components.interfaces.cellml_contextIModelListMonitor) ||
        aIID.equals(Components.interfaces.nsISupports))
      return this;
    return NULL;
  },

  destroy: function()
  {
    var i;
    for (i in this)
      delete this[i];
  },

  modelRenamed: function(renamedNode, newName)
  {
    var ol = this.rdfservice.GetLiteral(renamedNode.name);
    var nl = this.rdfservice.GetLiteral(newName);
    var rsrc = this.getResourceByNode(renamedNode);
    if (rsrc == null)
      return;
    // The final bit must be done on the main thread...
    var i;
    for (i = 0; i < 3; i++)
      this.sources[i].Change(rsrc, this.rr.name, ol, nl);
  },

  modelAdded: function(newNode, depth)
  {
    var l = newNode.parentList;
    if (l == null)
      return;
      
    // We will always insert at a certain point in the change sequence...
    var changeseq = this.getChangeSequenceByList(l);
    if (changeseq == null)
      return;
      
    // The history list is very simple: we just add the new resource at the
    // end...
    var historyseq = this.historySeq;
      
    // Find the toplevel object.
    var tla = newNode, tlb;
    while ((tlb = tla.parentList.parentNode) != null)
    {
      tla = tlb;
    }
    // This might return null, if this is a completely new model,
    // rather than a new derivative model.
    var oldLatest = this.getLatestResourceByNode(tla);
    var latestseq = this.latestSeq;

    // Manufacture a new resource...
    var rsrc = this.rdfservice.GetAnonymousResource();

    // Make it open...
    this.persist.Assert(rsrc, this.rr.open, this.rr.true, true);

    // Update nodeByResource and resourceByNode...
    this.nodeByResource[rsrc.ValueUTF8] = newNode;
    this.resourceByNode[
      newNode.QueryInterface(Components.interfaces.IWrappedPCM).objid] =
      rsrc;
    this.latestResourceByTopNode[
      tla.QueryInterface(Components.interfaces.IWrappedPCM).objid] =
      rsrc;
    // Produce a new sequence for the changes view...
    var seq = this.rdfcu.MakeSeq(this.sources[CHANGES_VIEW], rsrc);
    var dn = newNode.derivedModels;
    var objid = dn.QueryInterface(Components.interfaces.
                                  IWrappedPCM).objid;
    this.rdfChangeSequenceByList[objid] = seq;

    // Get the name literal of the new node...
    var name = this.rdfservice.GetLiteral(newNode.name);
    
    // Assert the name onto the new node...
    var i;
    for (i = 0; i < 3; i++)
      this.sources[i].Assert(rsrc, this.rr.name, name,
                             true);
    
    changeseq.AppendElement(rsrc);
    historyseq.AppendElement(rsrc);
    if (oldLatest != null)
    {
      var oldidx = latestseq.IndexOf(oldLatest);
      latestseq.RemoveElementAt(oldidx, false);
      latestseq.InsertElementAt(rsrc, oldidx, false);
    }
    else
      latestseq.AppendElement(rsrc);
    this.builder.rebuild();
    // I think it is useful to select the new model.
    var tb = this.builder.QueryInterface(Components.interfaces.
                                         nsIXULTreeBuilder);
    var idx = tb.getIndexOfResource(rsrc);
    this.view.selection.select(idx);
  },

  modelFrozenStateChanged: function(node, newState)
  {
    if (newState == false)
    {
      const IDataCollector = Components.interfaces.IDataCollector;
      UntagWithName(node.model, "data_collector");
      UntagWithName(node.model, "compiled_model");
      UntagWithName(node.model, "integration_run");
      var l = window.unfreezeListeners;
      if (l != null)
      {
        var ul;
        for (ul in l)
        {
          ul = l[ul];
          if (ul.currentModel == null)
            continue;
          if (CompareObjects(node, ul.currentModel))
            ul.unfrozen();
        }
      }
      window.modelSelector.callSelectHandlers();
    }
  },

  findLatestChildButOne: function(root, excludeid)
  {
    var l = root.derivedModels;
    if (l == null)
      return null;
    var it = l.iterateModelNodes();
    var nmn;
    // Get our timestamp...
    var ts = root.modificationTimestamp;
    var hit = root;
    var objid = root.QueryInterface(Components.interfaces.
                                    IWrappedPCM).objid;
    if (objid == excludeid)
    {
      return null;
    }
    
    var nmn;
    while ((nmn = it.nextModelNode()) != null)
    {
      var m = this.findLatestChildButOne(nmn, excludeid);
      if (m == null)
        continue;
      if (m[1] > ts)
      {
        hit = m[0];
        ts = m[1];
      }
    }
    return [hit, ts];
  },
  
  modelRemoved: function(oldNode, depth)
  {
    // If necessary, recursively remove children first (this ensures that
    //   we correctly compute things like the new latest node, although it
    //   is more expensive than what is required).
    var it = oldNode.derivedModels.iterateModelNodes();
    var nmn;
    while ((nmn = it.nextModelNode()) != null)
      this.modelRemoved(nmn, depth + 1);

    // Find the resource...
    var rsrc = this.getResourceByNode(oldNode);
    if (rsrc == null)
      return;
      
    // Find the parent sequence...
    var l = oldNode.parentList;
      
    if (l == null)
      return;
    var changeseq = this.getChangeSequenceByList(l);
    if (changeseq == null)
      return;
      
    // The history list is a straight delete...
    var historyseq = this.historySeq;
      
    // There are two cases for the latest list...
    //  1) The node is not the latest. In that case do nothing.
    //  2) The node is the latest. Replace it with the next newest
    //     surviving node.
    var tla = oldNode, tlb;
    while (l != null)
    {
      tlb = l.parentNode;
      if (tlb == null)
        break;
      tla = tlb;
      l = tla.parentList;
    }
    // This might return null, if this is a completely new model,
    // rather than a new derivative model.
    var latestseq = this.latestSeq;
    var latest = this.getLatestResourceByNode(tla);
    // latest can't be null unless our internal state is messed up, so I
    // am deliberately not adding a check for it.
    var rLatest = latest.ValueUTF8;
    var rOld = rsrc.ValueUTF8;
    var newRsrc = null;
    var delRsrcLatest = false;
    if (rLatest == rOld)
    {
      delRsrcLatest = true;
      var objidOld = this.nodeByResource[rOld].
        QueryInterface(Components.interfaces.IWrappedPCM).objid;
      var replacement = this.findLatestChildButOne(tla, objidOld);
      // Find the resource for the replacement...
      newRsrc = null;
      if (replacement != null)
      {
        newRsrc = this.getResourceByNode(replacement[0]);
      }
      
      this.latestResourceByTopNode[
        tla.QueryInterface(Components.interfaces.IWrappedPCM).objid] =
          newRsrc;
    }

    // Remove from the resourceByNode array so it disappears later.
    this.resourceByNode[oldNode.QueryInterface(Components.interfaces.
                                               IWrappedPCM).objid] = null;
    this.nodeByResource[rsrc.ValueUTF8] = null;

    changeseq.RemoveElement(rsrc, true);
    historyseq.RemoveElement(rsrc, true);
    if (delRsrcLatest)
    {
      if (newRsrc)
      {
        var idx = latestseq.IndexOf(rsrc);
        latestseq.RemoveElementAt(idx, false);
        latestseq.InsertElementAt(newRsrc, idx, false);
      }
      else
        latestseq.RemoveElement(rsrc, true);
    }
    this.builder.rebuild();
  },

  getResourceByNode: function(node)
  {
    var objid = node.QueryInterface(Components.interfaces.IWrappedPCM).objid;
    return this.resourceByNode[objid];
  },

  getChangeSequenceByList: function(list)
  {
    var objid = list.QueryInterface(Components.interfaces.IWrappedPCM).objid;
    return this.rdfChangeSequenceByList[objid];
  },

  getLatestResourceByNode: function(node)
  {
    var objid = node.QueryInterface(Components.interfaces.IWrappedPCM).
      objid;
    return this.latestResourceByTopNode[objid];
  }
}
