// There are two model modes...
// 1. MODELMODE_LINK_SELECTED will always use the currently selected model.
const MODELMODE_LINK_SELECTED = 1
// 2. MODELMODE_FIXED has a fixed model, which it always uses.
const MODELMODE_FIXED         = 2

function GetGlyphGetter(obj)
{
  return function()
  {
    var idcr = Components.interfaces.IDataToCanvasRequest;
    if (obj.drawType == 'L')
      return idcr.GLYPH_LINES;
    else if (obj.drawType == 'SQ')
      return idcr.GLYPH_SQUARES;
    else if (obj.drawType == 'C')
      return idcr.GLYPH_CIRCLES;
    else if (obj.drawType == 'DI')
      return idcr.GLYPH_DIAMONDS;
    else if (obj.drawType == 'D')
      return idcr.GLYPH_DOTS;
    // Shouldn't be reached...
    alert("Unknown glyph string; this is a bug.");
    return -1;
  }
}

function GetDrawColourGetter(obj)
{
  return function()
  {
    return obj.getColourString();
  }
}

function SearchImportsForURI(uri, impIt)
{
  var imp;
  while ((imp = impIt.nextImport()) != null)
  {
    var m = imp.importedModel;
    if (m == null)
      continue;
    if (m.base_uri.asText == uri)
      return m;
    return SearchImportsForURI(uri, m.imports.iterateImports());
  }
  return null;
}

function SearchModelsForURI(uri, modelList)
{
  var it = modelList.iterateModelNodes();
  var mn;
  while ((mn = it.nextModelNode()) != null)
  {
    if (mn.model.base_uri.asText == uri)
      return mn.model;
    // Search the derived models...
    var m;
    m = SearchModelsForURI(uri, mn.derivedModels);
    if (m != null)
      return m;

    // Search the imports...
    return SearchImportsForURI(uri, mn.model.imports.iterateImports());
  }
  return null;
}

function FindVariableByResource(node)
{
  var v = node.QueryInterface(
            Components.interfaces.nsIRDFResource).ValueUTF8;
  var hash = v.indexOf("#");
  var uri = v.substr(0, hash);
  var ident = v.substr(hash + 1);

  // XXX this is broken for imports. In the future, we could fix this by
  // referring to a simulation, and getting the model from that.

  var model = SearchModelsForURI(uri, window.context.loadedModels);
  
  var cci = model.allComponents.iterateComponents();
  var c;

  while ((c = cci.nextComponent()) != null)
  {
    var vi = c.variables.iterateVariables();
    var v;
    while ((v = vi.nextVariable()) != null)
      if (v.cmetaId == ident)
        return v;
  }

  alert("Warning: No match for variable cmeta:id referenced from RDF.");
  return null;
}

function GraphLineInformationFromRDF(ds, n)
{
  var hints = {};
  var b = "http://www.cellml.org/metadata/graphs/1.0#";
  var typeR = window.rdfservice.GetResource(b + "type");
  var glyphR = window.rdfservice.GetResource(b + "glyph");
  var colourR = window.rdfservice.GetResource(b + "colour");
  var xvarR = window.rdfservice.GetResource(b + "x-variable");
  var yvarR = window.rdfservice.GetResource(b + "y-variable");
  hints.colour = ds.GetTarget(n, colourR, true);
  hints.colour = hints.colour.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
  var gli = new GraphLineInformation(null, hints);

  var i;
  for (i = 0; i < document.styleSheets.length; i++)
  {
    if (document.styleSheets[i].href ==
        "chrome://pcenv/skin/graphkey.css")
      break;
  }
  var sheet = document.styleSheets[i];

  var p = "prop";
  for (i = 0; i < 10; i++)
    p = p + "0123456789ABCDEF"[parseInt(Math.random() * 16)];
  gli.properties = p;
  var r = "treechildren::-moz-tree-cell(" + p + ") { background-color: " +
          hints.colour + "; }";
  sheet.insertRule(r, sheet.cssRules.length);

  var lineR = window.rdfservice.GetResource(b + "line");
  var dotR = window.rdfservice.GetResource(b + "dot");
  var squareR = window.rdfservice.GetResource(b + "square");
  var circleR = window.rdfservice.GetResource(b + "circle");
  var diamondR = window.rdfservice.GetResource(b + "diamond");
  var noneR = window.rdfservice.GetResource(b + "nonstandard-none");

  if (ds.GetTarget(n, typeR, true).EqualsNode(lineR))
    gli.drawType = 'L';
  else
  {
    var glyph = ds.GetTarget(n, glyphR, true);
    if (glyph.EqualsNode(dotR))
      gli.drawType = 'D';
    else if (glyph.EqualsNode(squareR))
      gli.drawType = 'SQ';
    else if (glyph.EqualsNode(circleR))
      gli.drawType = 'C';
    else if (glpyh.EqualsNode(diamondR))
      gli.drawType = 'DI';
    else if (glyph.EqualsNode(noneR))
      gli.drawType = 'I';
  }

  var varR = window.rdfservice.GetResource("http://www.cellml.org/metadata/graphs/1.0#variable");

  var xv = ds.GetTarget(n, xvarR, true);
  var v = FindVariableByResource(ds.GetTarget(xv, varR, true));
  gli.x = v;
  gli.xvar = v.name;
  gli.xcomp = v.componentName;

  var yv = ds.GetTarget(n, yvarR, true);
  var v = FindVariableByResource(ds.GetTarget(yv, varR, true));
  gli.y = v;
  gli.yvar = v.name;
  gli.ycomp = v.componentName;

  return gli;
}

function GraphLineInformation(grapharea, hints)
{
  if (grapharea != null)
  grapharea.changeBegin();

  if (window.unfreezeListeners == null)
    window.unfreezeListeners = [];
  window.unfreezeListeners.push(this);

  if (hints == null)
    hints = {};
  if (window.gliUniq == null)
    window.gliUniq = 1;
  this.gliUniq = window.gliUniq;
  window.gliUniq++;

  this.rdfresource = window.rdfservice.GetAnonymousResource();
  this.grapharea = grapharea;
  if (grapharea != null)
    this.renderContext = grapharea.gcontext;
  if (hints.colour != null)
    this.setColour(hints.colour);
  else
    this.colour = parseInt(Math.random() * (1<<24));
  this.drawType = 'L';
  this.__defineGetter__("glyph", GetGlyphGetter(this));
  this.__defineGetter__("drawColour", GetDrawColourGetter(this));
  this.firstPoint = 0;

  this.x = this.y = null;
  this.modelMode = null;
  this.setLink();

  this.xVariableMin = -1.0;
  this.yVariableMin = -1.0;
  this.xVariableMax = 1.0;
  this.yVariableMax = 1.0;
  this.xMultiplier = 1.0;
  this.xOffset = 0.0;
  this.yMultiplier = 1.0;
  this.yOffset = 0.0;
  this.lastPoint = 0x3FFFFFFF;
  this.xVariable = null;
  this.yVariable = null;

  if (grapharea != null)
    grapharea.changeEnd();
}

GraphLineInformation.prototype =
{
  QueryInterface: function(aIID)
  {
    if (aIID.equals(Components.interfaces.IDataToCanvasRequest) ||
        aIID.equals(Components.interfaces.nsISupports))
      return this;
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  getColourString: function()
  {
    var c = this.colour;
    var cs = "";
    var i;
    for (i = 0; i < 6; i++)
    {
      cs = "0123456789abcdef"[c & 0xF] + cs;
      c >>= 4;
    }
    return "#" + cs;
  },

  setColour: function(c)
  {
    this.colour = parseInt(c.substr(1), 16);
  },

  unfrozen: function()
  {
    if (this.currentModel != null)
    {
      var dc = EnsureDataCollector(this.currentModel.model);
      dc.removeDataCollectorObserver(this.grapharea);
      dc.addDataCollectorObserver(this.grapharea);
    }
    this.switchModels(this.currentModel);
  },

  setLink: function()
  {
    if (this.modelMode == MODELMODE_LINK_SELECTED)
      return true;

    this.modelMode = MODELMODE_LINK_SELECTED;

    var cm = window.modelSelector.currentModel;
    if (cm == null)
      this.clearModel();
    else
      this.switchModels(cm);

    return true;
  },

  setDrawType: function(type)
  {
    this.drawType = type;
  },

  copySelectedModel: function()
  {
    this.grapharea.changeBegin();
    this.clearModel();
    this.modelMode = MODELMODE_FIXED;
    var cm = window.modelSelector.currentModel;
    if (cm == null)
    {
      alert("No model is currently selected.");
      this.grapharea.changeEnd();
      return false;
    }
    this.switchModels(cm);
    this.grapharea.changeEnd();

    return true;
  },

  // Internal API (externally use setLink and copySelectedModel).
  switchModels: function(model)
  {
    if (this.grapharea != null)
      this.grapharea.changeBegin();
    this.clearModel();
    this.currentModel = model;

    this.datacollector = EnsureDataCollector(this.currentModel.model);

    // See if variable changes...
    if (this.xcomp != null)
    {
      var c = this.currentModel.model.allComponents.
        getComponent(this.xcomp);
      if (c == null)
        this.x = null;
      else
      {
        var v = c.variables.getVariable(this.xvar);
        if (v == null)
          this.x = null;
        else
          this.x = v;
      }
    }
    if (this.ycomp != null)
    {
      var c = this.currentModel.model.allComponents.
        getComponent(this.ycomp);
      if (c == null)
        this.y = null;
      else
      {
        var v = c.variables.getVariable(this.yvar);
        if (v == null)
          this.y = null;
        else
          this.y = v;
      }
    }

    // Now set the X and Y labels...
    if (this.xvar != null)
    {
      if (this.x == null)
        this.xlabel.setAttribute("label", this.xvar + " (not found)");
      else
        this.xlabel.setAttribute("label", this.xvar + " (" + this.x.unitsName +
                                 ")");
    }
    if (this.yvar != null)
    {
      if (this.y == null)
        this.ylabel.setAttribute("label", this.yvar + " (not found)");
      else
        this.ylabel.setAttribute("label", this.yvar + " (" + this.y.unitsName +
                                 ")");
    }

    this.recheckVariables();

    if (this.grapharea != null)
    {
      this.grapharea.addModelLine(this.currentModel.model, this);
      this.grapharea.changeEnd();
    }
  },

  clearModel: function()
  {
    if (this.currentModel == null)
      return;

    this.grapharea.changeBegin();

    this.grapharea.removeModelLine(this.currentModel.model, this);

    EnsureDataCollector(this.currentModel.model).
      removeDataToCanvasRequest(this);

    // TODO: Deregister model handlers.
    this.currentModel = null;

    this.grapharea.changeEnd();
  },

  recheckVariables: function()
  {
    this.xVariable = this.yVariable = null;
    if (this.currentModel == null)
      return;
    var cm = GetTag(this.currentModel.model, "compiled_model");
    if (cm == null)
      return;
    cm = cm.compModel;
    this.integrationStarted(0, 0, 0, cm);
  },

  // External must call before abandaning this GraphLineInformation...
  deregister: function()
  {
    var i;
    for (i in window.unfreezeListeners)
      if (window.unfreezeListeners[i] == this)
        delete window.unfreezeListeners[i];
    this.clearModel();
  },

  // Callback hook, only to be called from the graph area.
  selected: function(newModel)
  {
    if (this.modelMode == MODELMODE_LINK_SELECTED)
      this.switchModels(newModel);
  },

  SetScale: function(xMin, xMax, yMin, yMax, mx, cx, my, cy)
  {
    this.xVariableMin = xMin;
    this.yVariableMin = yMin;
    this.xVariableMax = xMax;
    this.yVariableMax = yMax;
    this.xMultiplier = mx;
    this.xOffset = cx;
    this.yMultiplier = my;
    this.yOffset = cy;
  },

  integrationStarted: function(start, stop, increment, cmodel)
  {
    if (!CompareObjects(cmodel.model, this.currentModel.model))
      return;
    var ci = cmodel.codeInformation;
    var vi = ci.iterateVariables();
    var v;
    while ((v = vi.nextVariable()) != null)
    {
      if (CompareObjects(v.source, this.x))
      {
        if (v.type == 0/*BOUND*/)
          this.xVariable = ci.variableCount;
        else if (v.type == 1 || v.type == 2)
          return; // ignore the variable, it doesn't vary.
        else
          this.xVariable = v.variableIndex;
      }
      if (CompareObjects(v.source, this.y))
      {
        if (v.type == 0/*BOUND*/)
          this.yVariable = ci.variableCount;
        else if (v.type == 1 || v.type == 2)
          return; // ignore the variable, it doesn't vary.
        else
          this.yVariable = v.variableIndex;
      }
    }
  },

  enabled: function()
  {
    if (this.currentModel == null ||
        this.x == null || this.y == null ||
        this.drawType == 'I' || this.xVariable == null ||
        this.yVariable == null)
      return false;
    return true;
  },

  getDrawTypeResource: function()
  {
    var b = "http://www.cellml.org/metadata/graphs/1.0#";
    if (this.drawType != 'L')
      return window.rdfservice.GetResource(b + "scatter");
    return window.rdfservice.GetResource(b + "line");
  },

  getGlyphTypeResource: function()
  {
    var b = "http://www.cellml.org/metadata/graphs/1.0#";
    var suf;
    if (this.drawType == 'D')
      suf = "dot";
    else if (this.drawType == 'SQ')
      suf = "square";
    else if (this.drawType == 'C')
      suf = "circle";
    else if (this.drawType == 'DI')
      suf = "diamond";
    else
      suf = "nonstandard-none";
    return window.rdfservice.GetResource(b + suf);
  },

  saveSessionState: function(ds, n)
  {
    var b = "http://www.cellml.org/metadata/graphs/1.0#";
    var typeR = window.rdfservice.GetResource(b + "type");
    var glyphR = window.rdfservice.GetResource(b + "glyph");
    var colourR = window.rdfservice.GetResource(b + "colour");
    var xvarR = window.rdfservice.GetResource(b + "x-variable");
    var yvarR = window.rdfservice.GetResource(b + "y-variable");

    ds.Assert(n, typeR, this.getDrawTypeResource(), true);
    if (this.drawType != 'L')
      ds.Assert(n, glyphR, this.getGlyphTypeResource(), true);

    ds.Assert(n, colourR, window.rdfservice.GetLiteral(this.getColourString()),
              true);

    var xr = window.rdfservice.GetAnonymousResource();
    ds.Assert(n, xvarR, xr, true);
    var varR = window.rdfservice.GetResource("http://www.cellml.org/metadata/graphs/1.0#variable");

    if (this.x != null)
    {
      if (this.x.cmetaId == "")
        this.x.cmetaId = this.x.componentName + "_" +
          this.x.name;
      var uri = this.x.modelElement.base_uri.asText + "#" +
        this.x.cmetaId;
      ds.Assert(xr, varR, window.rdfservice.GetResource(uri), true);
    }

    var yr = window.rdfservice.GetAnonymousResource();
    ds.Assert(n, yvarR, yr, true);

    if (this.y != null)
    {
      if (this.y.cmetaId == "")
        this.y.cmetaId = this.y.componentName + "_" +
          this.y.name;
      var uri = this.y.modelElement.base_uri.asText + "#" +
        this.y.cmetaId;
      ds.Assert(yr, varR, window.rdfservice.GetResource(uri), true);
    }
  }
};

function GraphAxis(name, owner)
{
  this.type = "linegraph";
  this.name = name;
  this.locked = true;
  this.lockedBy = owner;
  this.lines = [];
  this.resource = window.rdfservice.GetAnonymousResource();
}

function GraphAxisFromRDF(ds, res)
{
  var nameR = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/graphname");
  var firstR = window.rdfservice.GetResource(
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#first");
  var restR = window.rdfservice.GetResource(
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest");
  var nilR = window.rdfservice.GetResource(
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil");
  var urlR = window.rdfservice.GetResource(
    "http://www.cellml.org/tools/pcenv/externalurl");
  var name = ds.GetTarget(res, nameR, true);
  name = name.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;

  var a = new GraphAxis(name, null);
  a.resource = res;
  a.locked = false;

  try
  {
    var url = ds.GetTarget(res, urlR, true);
    if (url != null)
    {
      a.type = "content";
      a.contentURL = url.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
      return a;
    }
  }
  catch (e)
  {
    alert(e);
  }

  var traceR = window.rdfservice.GetResource(
    "http://www.cellml.org/metadata/graphs/1.0#traces");
  var tl = ds.GetTarget(res, traceR, true);

  tl = ds.GetTarget(tl, firstR, true);
  if (tl == null || tl.EqualsNode(nilR))
    return a;

  while ((tl != null) && !tl.EqualsNode(nilR))
  {
    var n = ds.GetTarget(tl, firstR, true);

    a.lines.push(GraphLineInformationFromRDF(ds, n));
    tl = ds.GetTarget(tl, restR, true);
  }

  return a;
}

GraphAxis.prototype = {
  deregister: function() {
    var i;
    for (i in this.lines)
      this.lines[i].deregister();
    this.lines = null;
  },

  saveSessionState: function(ds) {
    var nameR = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/graphname");
    ds.Assert(this.resource, nameR, window.rdfservice.GetLiteral(this.name), true);
    var traceR = window.rdfservice.GetResource(
      "http://www.cellml.org/metadata/graphs/1.0#traces");
    var firstR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#first");
    var restR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest");
    var nilR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil");
    var urlR = window.rdfservice.GetResource(
      "http://www.cellml.org/tools/pcenv/externalurl");

    if (this.type == 'content')
    {
      ds.Assert(this.resource, urlR,
                window.rdfservice.GetLiteral(this.contentURL), true);
      return;
    }

    var trace = window.rdfservice.GetAnonymousResource();
    ds.Assert(this.resource, traceR, trace, true);

    // This case is a bit different...
    if (this.lines.length == 0)
    {
      ds.Assert(trace, firstR, nilR, true);
      return;
    }

    var l;
    var first = true;
    for (l in this.lines)
    {
      var tmp = window.rdfservice.GetAnonymousResource();
      ds.Assert(trace, first ? firstR : restR, tmp, true);
      trace = tmp;

      if (first)
        first = false;

      var r = window.rdfservice.GetAnonymousResource();
      ds.Assert(trace, firstR, r, true);
      this.lines[l].saveSessionState(ds, r);
    }

    ds.Assert(trace, restR, nilR, true);
  }
};

function GraphAxisSet()
{
  this.mObs = [];
  this.mAxes = [];
  this.mLastAxid = 0;
  this.needUpdate = false;
  this.inChanges = 0;
}

GraphAxisSet.prototype = {
  addObserver: function(obs) {
    this.mObs.push(obs);
  },

  removeObserver: function(obs) {
    var i;
    for (i in this.mObs)
      if (this.mObs[i] == obs)
        delete this.mObs[i];
    if (this.mObs.length == 0)
    {
      for (i in this.mAxes)
        this.mAxes[i].deregister();
      this.mAxes = [];
    }
  },

  addAxis: function(axis) {
    this.mLastAxid = axis.axid = this.mLastAxid + 1;
    this.mAxes[axis.axid] = axis;
    this.beginChange();
    this.needUpdate = true;
    this.endChange();
  },

  deleteAxis: function(axis) {
    delete this.mAxes[axis.axid];
    axis.deregister();
    this.beginChange();
    this.needUpdate = true;
    this.endChange();
  },

  sessionAxisAvailable: function(graph, resource) {
    // Find the axis...
    var i;
    for (i in this.mAxes)
    {
      if (this.mAxes[i].resource.EqualsNode(resource))
      {
        var ax = this.mAxes[i];
        var j;
        for (j in ax.lines)
          ax.lines[j].recheckVariables();
        graph.setAxis(i);
        return;
      }
    }
    alert("Warning: No match for session axis found.");
  },
 
  lockAxis: function(axid, lockedBy, oldAxid) {
    var axis = this.mAxes[axid];
    if (axis.locked)
    {
      // A straight swap...
      axis.lockedBy.setAxis(oldAxid);
    }

    var l;
    for (l in axis.lines)
    {
      l = axis.lines[l];
      if (l.currentModel != null && l.grapharea != null)
      {
        l.grapharea.removeModelLine(l.currentModel.model, l);
      }
      l.grapharea = lockedBy.grapharea;
      if (l.currentModel != null)
      {
        l.grapharea.addModelLine(l.currentModel.model, l);
      }
      l.renderContext = l.grapharea.gcontext;
    }

    axis.lockedBy = lockedBy;
    axis.locked = true;
    this.beginChange();
    this.needUpdate = true;
    this.endChange();
    return axis;
  },

  beginChange: function() {
    this.inChanges++;
  },

  endChange: function() {
    this.inChanges--;
    if (this.inChanges == 0 && this.needUpdate)
    {
      this.dispatchUpdate();
      this.needUpdate = false;
    }
  },

  unlockAxis: function(axis) {
    axis.locked = false;
    axis.lockedBy = null;
    this.beginChange();
    this.needUpdate = true;
    this.endChange();
  },

  dispatchUpdate: function() {
    for (i in this.mObs)
      this.mObs[i].update(this.mAxes);
  },

  renameAxis: function(axis, name) {
    this.mAxes[axis.axid].name = name;
    this.dispatchUpdate();
  },

  saveSessionState: function(ds) {
    var firstR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#first");
    var restR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest");
    var nilR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil");

    // Non-standard top-level...
    var tl = window.rdfservice.GetResource("http://www.cellml.org/tools/" +
                                           "pcenv/displayGraphs");
    if (this.mAxes.length == 0)
    {
      ds.Assert(tl, firstR, nilR, true);
      return;
    }

    var first = true;

    for (g in this.mAxes)
    {
      var tmp = window.rdfservice.GetAnonymousResource();
      ds.Assert(tl, first ? firstR : restR, tmp, true);
      tl = tmp;

      if (first)
        first = false;

      this.mAxes[g].saveSessionState(ds);
      ds.Assert(tl, firstR, this.mAxes[g].resource, true);
    }

    ds.Assert(tl, restR, nilR, true);
  },

  loadSessionState: function(ds)
  {
    var firstR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#first");
    var restR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest");
    var nilR = window.rdfservice.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil");
    var tl = window.rdfservice.GetResource("http://www.cellml.org/tools/" +
                                           "pcenv/displayGraphs");

    this.mAxes = [];
    this.mLastAxid = 0;
    
    try
    {
      tl = ds.GetTarget(tl, firstR, true);
    }
    catch (e)
    {
      alert(e);
      return;
    }

    while (!tl.EqualsNode(nilR))
    {
      var n = ds.GetTarget(tl, firstR, true);
      var axis = GraphAxisFromRDF(ds, n);
      this.mLastAxid = axis.axid = this.mLastAxid + 1;
      this.mAxes[axis.axid] = axis;
      
      try
      {
        tl = ds.GetTarget(tl, restR, true);
      }
      catch (e)
      {
        alert(e);
        return;
      }
    }
  }
};
