function LoadCompleted()
{
  window.loadsInProgress--;

  if (window.loadsInProgress == 0)
  {
    if (window.allLoadsComplete != null)
    {
      var i;
      for (i in window.allLoadsComplete)
        window.allLoadsComplete[i]();
      window.allLoadsComplete = null;
    }
  }
}

function ImportLoadCompleted(r, imp, model, importURL, executeOn)
{
  if (IsTagged(model, "load_failed"))
    return;
  UntagWithName(model, "import_in_progress");
  var rt = r.responseText;

  if (imp.wasInstantiated)
  {
    return;
  }
  try
  {
    imp.instantiateFromText(rt);
  }
  catch (e)
  {
    Status("Import " + importURL + " parse failed. Aborting model load.");
    TagWithName(model, "load_failed");
    return;
  }
  Status("Import " + importURL + " completed. Looking for further " +
         "imports...");
  var rt = r;
  if (rt != null)
    rt = rt.toplevel;
  GetMoreImportsOrExecute(rt, importURL, model, executeOn);
}

function ImportLoadFailed(r, imp, model, importURL)
{
  Status("Load of import " + importURL + " failed. Model load cancelled.");
  UntagWithName(model, "import_in_progress");
  TagWithName(model, "load_failed");
}

function AttemptAbsoluteURL(newURL, topURL)
{
  if (newURL.indexOf("://") > 0)
  {
    return newURL;
  }
  var protIdx = topURL.indexOf("://");
  var protEnd;
  if (protIdx < 0)
    protEnd = 0;
  else
    protEnd = protIdx + 3;
  var pathIdx = topURL.indexOf("/", protEnd);

  // If no protocol and no path, we might as well return the original string...
  if (protIdx < 0 && pathIdx < 0)
  {
    return newURL;
  }

  if (newURL[0] == "/")
  {
    // newURL is /foo/bar/test.xml, topURL is e.g. http://www.example.org
    if (pathIdx < 0)
    {
      return topURL + newURL;
    }
    return topURL.substr(0, pathIdx) + newURL
  }
  
  if (pathIdx < 0)
    return topURL + "/" + newURL;

  var lastSlash = topURL.lastIndexOf("/");
  var cand = topURL.substr(0, lastSlash);

  // Now try stripping ../ off beginning of newURL...
  while (true)
  {
    if (newURL[0] != '.')
      break;
    if (newURL[1] == '/')
    {
      newURL = newURL.substr(2);
      continue;
    }
    if (newURL[1] != '.')
      break;
    if (newURL[2] != '/')
      break;
    lastSlash = cand.lastIndexOf("/");
    if (lastSlash == -1)
      return cand + newURL;
    newURL = newURL.substr(3);
    cand = cand.substr(0, lastSlash);
  }
  return cand + "/" + newURL;
}

function GetImportLoadCompleted(imp, model, URL, executeOn)
{
  return function (event){ ImportLoadCompleted(event.target, imp, model, URL, executeOn); };
}

function GetImportLoadFailed(imp, model, URL)
{
  return function (event){ ImportLoadFailed(event.target, imp, model, URL); };
}

function RecursivelyFindUninstantiatedImports(rt, modelURL, model, importSet,
                                              executeOn)
{
  var cii = importSet.iterateImports();
  var allDone = true;
  var imp;
  while ((imp = cii.nextImport()) != null)
  {
    if (!imp.wasInstantiated &&
        !IsTagged(imp, "import_in_progress"))
    {
      allDone = false;
      TagWithName(imp, "import_in_progress");
      var URL = imp.xlinkHref.asText;
      var r = new XMLHttpRequest();
      r.toplevel = rt;
      URL = AttemptAbsoluteURL(URL, modelURL);
      Status("Attempting to load URL " + URL);
      r.onload = GetImportLoadCompleted(imp, model, URL, executeOn);
      r.onerror = GetImportLoadFailed(imp, model, URL);
      r.open("GET", URL, true);
      try
      {
        r.send(null);
      }
      catch (e)
      {
        ImportLoadFailed({target: r}, imp, modelURL, URL);
      }
    }
    else
    {
      // Import already loaded, so visit its children...
      var im = imp.importedModel;
      if (im != null)
      {
        allDone &= RecursivelyFindUninstantiatedImports(rt, modelURL, model, im.imports,
                                                        executeOn);
      }
      else
      {
        allDone = false;
      }
    }
  }
  return allDone;
}

window.ExecuteRegistration = {
  loadComplete: function(r, modelURL, model)
  {
    var c = window.context;
    if (c == null)
    {
      c = GetContext();
    }

    var rootTo = c.loadedModels;

    // Now it is time to see if we can root the model...
    if (r.info != null && r.info.rootTo != null)
    {
      if (r.info.rootTo.isRooted)
        rootTo = r.info.rootTo.isRooted;
      else
      {
        // Can't root the model. Tell the parent that it has children waiting
        // to be rooted...
        var p = r.info.rootTo;
        if (p.wantRoot == null)
          p.wantRoot = [];
        p.wantRoot.push({req: r, modelURL: modelURL, model: model});
        return;
      }
    }

    var modelName = model.cmetaId;
    if (modelName == "")
      modelName = modelURL;
    if (r != null && r.info != null)
      modelName = r.info.name;

    Status("Context-side model object created. Registering...");
    var c = window.context;
    if (c == null)
    {
      c = GetContext();
    }

    var n = rootTo.makeNode(model);
    n.name = modelName;

    // XXX do we need an r.info override option for the modified date?
    n.stampModifiedNow();

    if (r != null && r.info != null)
      n.isFrozen = r.info.freeze;

    rootTo.addModel(n);
    Status("Model " + modelName + " loaded.");

    // Any children which get rooted in the future will need our derived models
    // list...
    r.isRooted = n.derivedModels;
    n = null;

    // Now that the parent is rooted, we can root any children waiting to be
    // rooted...
    var i;
    for (i in r.wantRoot)
    {
      var ch = r.wantRoot[i];
      this.loadComplete(ch.req, ch.modelURL, ch.model);
    }
    LoadCompleted();
  }
};

function GetMoreImportsOrExecute(r, modelURL, model, executeOn)
{
  if (!RecursivelyFindUninstantiatedImports(r, modelURL, model, model.imports, executeOn))
    return;

  executeOn.loadComplete(r, modelURL, model);
}

function ModelLoadCompleted(r, modelURL)
{
  Status("Model XML loaded, transferring to your CellML context...");
  if (r != null && r.responseXML != null)
  {
    var ns = r.responseXML.documentElement.namespaceURI;
    if (ns == "http://www.cellml.org/cellml/1.0#" ||
        ns == "http://www.cellml.org/cellml/1.1#")
      ;
    else
    {
      Status("Model is not CellML!");
      if (ns == "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
      {
        if (confirm("The file you attempted to load is an RDF document and " +
                    "not a CellML model. It may be a PCEnv session file.\n" +
                    "Would you like to load the file as a session?\n" +
                    "Warning: You will lose any unsaved changes to models or " +
                    "simulation parameters, and any graphs that you have set up.",
                    "Replace current session?"))
          ReadSessionFromURL(modelURL);
      }
      else
      {
        alert("Sorry, the loaded document does not contain a CellML model.");
      }
      LoadCompleted();
      return;
    }
  }
  var c = window.context;
  if (c == null)
  {
    c = GetContext();
  }

  var ml = c.cellmlBootstrap.modelLoader;
  var m;
  try
  {
    m = ml.createFromText(r.responseText);
  }
  catch (e)
  {
    Status("Model load failed: " + ml.lastErrorMessage);
    return;
  }

  m.base_uri.asText = modelURL;

  GetMoreImportsOrExecute(r, modelURL, m, window.ExecuteRegistration);
}

function ModelLoadError(event)
{
  Status("Model load failed.");
  LoadCompleted();
}

function GetModelLoadCompleted(url)
{
  return function(event){ ModelLoadCompleted(event.target, url);};
}

function OpenHTTPRequest(url, info)
{
  var r = new XMLHttpRequest();
  r.toplevel = r;
  r.info = info;
  r.open("GET", url);
  r.onload = GetModelLoadCompleted(url);
  r.onerror = ModelLoadError;
  try
  {
    if (window.loadsInProgress == null)
      window.loadsInProgress = 0;
    window.loadsInProgress++;
    r.send(null);
    return r;
  }
  catch (e)
  {
    ModelLoadError({target: r});
  }
  return null;
}
