function SaveSplitterState(ds, node, element)
{
  if (element.getAttribute("state") == "collapsed")
  {
    var t = window.rdfservice.GetLiteral("true");
    if (element.getAttribute("substate") == "before")
    {
      var cb = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_collapsedBefore");
      ds.Assert(node, cb, t, true);
    }
    else
    {
      var ca = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_collapsedAfter");
      ds.Assert(node, ca, t, true);
    }
  }
  if (element.previousSibling.hasAttribute("height"))
  {
    var ph = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_prevHeight");
    var lit = window.rdfservice.GetLiteral(element.previousSibling.getAttribute("height"));
    ds.Assert(node, ph, lit, true);
  }
  if (element.nextSibling.hasAttribute("height"))
  {
    var nh = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_nextHeight");
    var lit = window.rdfservice.GetLiteral(element.nextSibling.getAttribute("height"));
    ds.Assert(node, nh, lit, true);
  }
  if (element.previousSibling.hasAttribute("width"))
  {
    var ph = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_prevWidth");
    var lit = window.rdfservice.GetLiteral(element.previousSibling.getAttribute("width"));
    ds.Assert(node, ph, lit, true);
  }
  if (element.nextSibling.hasAttribute("width"))
  {
    var nh = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_nextWidth");
    var lit = window.rdfservice.GetLiteral(element.nextSibling.getAttribute("width"));
    ds.Assert(node, nh, lit, true);
  }
}

function LoadSplitterState(ds, node, element)
{
  var t = window.rdfservice.GetLiteral("true");
  var cb = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_collapsedBefore");
  var ca = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_collapsedAfter");

  var isCB = ds.HasAssertion(node, cb, t, true);
  var isCA = ds.HasAssertion(node, ca, t, true);

  if (isCB || isCA)
  {
    if (isCB)
      element.setAttribute("substate", "before");
    else
      element.setAttribute("substate", "after");
    element.setAttribute("state", "collapsed");
  }
  else
    element.removeAttribute("state");

  var tab = [
      {
        pred: window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_prevHeight"),
        sib: element.prevSibling,
        att: "height"
      },
      {
        pred: window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_nextHeight"),
        sib: element.nextSibling,
        att: "height"
      },
      {
        pred: window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_prevWidth"),
        sib: element.prevSibling,
        att: "width"
      },
      {
        pred: window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/splitterState_nextWidth"),
        sib: element.nextSibling,
        att: "width"
      }
            ];

  var i;
  for (i in tab)
  {
    var e = tab[i];
    try
    {
      var l = ds.GetTarget(node, e.pred, true);
      l = l.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
      e.sib.setAttribute(e.att, l);
    }
    catch (ex)
    {
    }
  }
}

function WriteSessionToDataSource(ds)
{
  var mls = document.getElementById("modlayoutsplit");
  var ssN = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/modlayoutsplit");
  SaveSplitterState(ds, ssN, mls);
  window.modelSelector.saveSessionState(ds);
  document.getElementById("layout").saveSessionState(ds);
  window.axisSet.saveSessionState(ds);
}

function ReadSessionFromDataSource(ds)
{
  var mls = document.getElementById("modlayoutsplit");
  var ssN = window.rdfservice.GetResource("http://www.cellml.org/tools/pcenv/modlayoutsplit");
  LoadSplitterState(ds, ssN, mls);
  window.modelSelector.loadSessionState(ds);

  var f = function() {
    window.axisSet.beginChange();
    window.axisSet.loadSessionState(ds);
    document.getElementById("layout").loadSessionState(ds);
    window.axisSet.endChange();
  }

  if (window.loadsInProgress == null || window.loadsInProgress == 0)
    f();
  else
  {
    if (window.allLoadsComplete == null)
      window.allLoadsComplete = [];
    window.allLoadsComplete.push(f);
  }
}

function WriteSessionTo(file)
{
  // Create an in-memory RDF datastore...
  var c = Components.classes["@mozilla.org/rdf/datasource;1?name=xml-" +
                             "datasource"];
  var ifce = Components.interfaces.nsIRDFDataSource;
  var ds = c.createInstance(ifce);
  var dsr = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
  var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
  var url = ioService.newFileURI(file).spec;
  dsr.Init(url);

  WriteSessionToDataSource(ds);

  dsr.Flush();
  window.rdfservice.UnregisterDataSource(ds);
}

function ReadSessionFromURL(url)
{

  // Create an in-memory RDF datastore...
  var c = Components.classes["@mozilla.org/rdf/datasource;1?name=xml-" +
                             "datasource"];
  var ifce = Components.interfaces.nsIRDFDataSource;
  var ds = c.createInstance(ifce);
  var dsr = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
  dsr.Init(url);
  dsr.Refresh(true);

  ReadSessionFromDataSource(ds);

  window.rdfservice.UnregisterDataSource(ds);
}

function ReadSessionFrom(file)
{
  var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
  var url = ioService.newFileURI(file).spec;
  ReadSessionFromURL(url);
}
